# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Iterable

from pants.backend.shell.target_types import ShellLibrary, Shunit2Tests, Shunit2TestsSources
from pants.core.goals.tailor import (
    AllOwnedSources,
    PutativeTarget,
    PutativeTargets,
    PutativeTargetsRequest,
    group_by_dir,
)
from pants.engine.fs import PathGlobs, Paths
from pants.engine.internals.selectors import Get
from pants.engine.rules import collect_rules, rule
from pants.engine.target import Target
from pants.engine.unions import UnionRule
from pants.source.filespec import Filespec, matches_filespec
from pants.util.logging import LogLevel


@dataclass(frozen=True)
class PutativeShellTargetsRequest:
    pass


def classify_source_files(paths: Iterable[str]) -> dict[type[Target], set[str]]:
    """Returns a dict of target type -> files that belong to targets of that type."""
    tests_filespec = Filespec(includes=list(Shunit2TestsSources.default))
    test_filenames = set(
        matches_filespec(tests_filespec, paths=[os.path.basename(path) for path in paths])
    )
    test_files = {path for path in paths if os.path.basename(path) in test_filenames}
    library_files = set(paths) - test_files
    return {Shunit2Tests: test_files, ShellLibrary: library_files}


@rule(level=LogLevel.DEBUG, desc="Determine candidate shell targets to create")
async def find_putative_targets(
    _: PutativeShellTargetsRequest, all_owned_sources: AllOwnedSources
) -> PutativeTargets:
    all_shell_files = await Get(Paths, PathGlobs(["**/*.sh"]))
    unowned_shell_files = set(all_shell_files.files) - set(all_owned_sources)
    classified_unowned_shell_files = classify_source_files(unowned_shell_files)
    pts = []
    for tgt_type, paths in classified_unowned_shell_files.items():
        for dirname, filenames in group_by_dir(paths).items():
            name = "tests" if tgt_type == Shunit2Tests else os.path.basename(dirname)
            kwargs = {"name": name} if tgt_type == Shunit2Tests else {}
            pts.append(
                PutativeTarget.for_target_type(
                    tgt_type, dirname, name, sorted(filenames), kwargs=kwargs
                )
            )
    return PutativeTargets(pts)


def rules():
    return [*collect_rules(), UnionRule(PutativeTargetsRequest, PutativeShellTargetsRequest)]
